import asyncio
from typing import Optional

from nats.aio import subscription
from nats.js import api, client

from nats_app.meta import PullSubscriptionMeta, PushSubscriptionMeta, SubscriptionMeta


class NATSRouter:
    push_subscribers: list[SubscriptionMeta]
    js_push_subscribers: list[PushSubscriptionMeta]
    js_pull_subscribers: list[PullSubscriptionMeta]
    prefix: str = ""

    def __init__(self, prefix: str = ""):
        self.push_subscribers = []
        self.js_push_subscribers = []
        self.js_pull_subscribers = []
        self.prefix = prefix

    def push_subscribe(
        self,
        subject,
        queue: str = "",
        future: Optional[asyncio.Future] = None,
        max_msgs: int = 0,
        pending_msgs_limit: int = subscription.DEFAULT_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: int = subscription.DEFAULT_SUB_PENDING_BYTES_LIMIT,
        by_pass_msg: Optional[str] = False,
    ):
        def wrapper(fn):
            self.push_subscribers.append(
                SubscriptionMeta(
                    subject=subject,
                    queue=queue,
                    handler=fn,
                    future=future,
                    max_msgs=max_msgs,
                    pending_msgs_limit=pending_msgs_limit,
                    pending_bytes_limit=pending_bytes_limit,
                )
            )

        return wrapper

    def js_push_subscribe(
        self,
        subject: str,
        queue: Optional[str] = None,
        durable: Optional[str] = None,
        stream: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        manual_ack: bool = False,
        ordered_consumer: bool = False,
        idle_heartbeat: Optional[float] = None,
        flow_control: bool = False,
        pending_msgs_limit: int = client.DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: int = client.DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
        deliver_policy: Optional[api.DeliverPolicy] = None,
        headers_only: Optional[bool] = None,
        inactive_threshold: Optional[float] = None,
    ):
        def wrapper(fn):
            self.js_push_subscribers.append(
                PushSubscriptionMeta(
                    subject=subject,
                    queue=queue,
                    cb=fn,
                    durable=durable,
                    stream=stream,
                    config=config,
                    manual_ack=manual_ack,
                    ordered_consumer=ordered_consumer,
                    idle_heartbeat=idle_heartbeat,
                    flow_control=flow_control,
                    pending_msgs_limit=pending_msgs_limit,
                    pending_bytes_limit=pending_bytes_limit,
                    deliver_policy=deliver_policy,
                    headers_only=headers_only,
                    inactive_threshold=inactive_threshold,
                )
            )

        return wrapper

    def js_pull_subscribe(
        self,
        subject: str,
        durable: Optional[str] = None,
        stream: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        pending_msgs_limit: int = client.DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: int = client.DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
        inbox_prefix: bytes = api.INBOX_PREFIX,
        batch: int = 1,
        timeout: Optional[float] = None,
        heartbeat: Optional[float] = None,
    ):
        def wrapper(fn):
            self.js_pull_subscribers.append(
                PullSubscriptionMeta(
                    handler=fn,
                    subject=subject,
                    durable=durable,
                    stream=stream,
                    config=config,
                    pending_msgs_limit=pending_msgs_limit,
                    pending_bytes_limit=pending_bytes_limit,
                    inbox_prefix=inbox_prefix,
                    batch=batch,
                    timeout=timeout,
                    heartbeat=heartbeat,
                )
            )

        return wrapper

    def include_router(self, router):
        for m in router.push_subscribers:
            self.push_subscribers.append(
                SubscriptionMeta(
                    subject=self.prefix + m.subject,
                    queue=m.queue,
                    handler=m.handler,
                    future=m.future,
                    max_msgs=m.max_msgs,
                    pending_msgs_limit=m.pending_msgs_limit,
                    pending_bytes_limit=m.pending_bytes_limit,
                )
            )

        for m in router.js_push_subscribers:
            self.js_push_subscribers.append(
                PushSubscriptionMeta(
                    subject=self.prefix + m.subject,
                    queue=m.queue,
                    cb=m.fn,
                    durable=m.durable,
                    stream=m.stream,
                    config=m.config,
                    manual_ack=m.manual_ack,
                    ordered_consumer=m.ordered_consumer,
                    idle_heartbeat=m.idle_heartbeat,
                    flow_control=m.flow_control,
                    pending_msgs_limit=m.pending_msgs_limit,
                    pending_bytes_limit=m.pending_bytes_limit,
                    deliver_policy=m.deliver_policy,
                    headers_only=m.headers_only,
                    inactive_threshold=m.inactive_threshold,
                )
            )

        for m in router.js_pull_subscribers:
            self.js_pull_subscribers.append(
                PullSubscriptionMeta(
                    subject=self.prefix + m.subject,
                    handler=m.fn,
                    durable=m.durable,
                    stream=m.stream,
                    config=m.config,
                    pending_msgs_limit=m.pending_msgs_limit,
                    pending_bytes_limit=m.pending_bytes_limit,
                    inbox_prefix=m.inbox_prefix,
                    batch=m.batch,
                    timeout=m.timeout,
                    heartbeat=m.heartbeat,
                )
            )



import pytest
from nats.js import JetStreamContext
from nats.js.api import StreamConfig
from nats.js.errors import NotFoundError

from nats_app.app import NATSApp


@pytest.mark.parametrize(
    "exist, new, fields, expected",
    [
        ({"a": 1, "b": 2}, {"a": 1, "b": 2}, ["a", "b"], True),
        ({"a": 1, "b": 2}, {"a": 1, "b": 3}, ["a", "b"], False),
        ({"a": 1, "b": 2}, {"a": 1, "b": None}, ["a", "b"], True),
        ({"a": 1, "b": 2}, {"a": 1}, ["a", "b"], True),
        ({"a": 1, "b": 2}, {"a": 2, "b": 2}, ["a", "b"], False),
    ],
)
def test_is_equal(exist, new, fields, expected):
    assert NATSApp._is_equal(exist, new, fields) == expected


@pytest.mark.parametrize(
    "exist, new, fields, expected",
    [
        (
            {"a": 1, "b": 2},
            {"a": 1, "b": 3},
            ["a", "b"],
            {"b": {"old": 2, "new": 3}},
        ),
        (
            {"a": 1, "b": 2},
            {"a": 1, "b": None},
            ["a", "b"],
            {},
        ),
        (
            {"a": 1, "b": 2},
            {"a": 2, "b": 2},
            ["a", "b"],
            {"a": {"old": 1, "new": 2}},
        ),
        (
            {"a": 1, "b": 2},
            {"a": 1, "b": 2, "c": 3},
            ["a", "b", "c"],
            {"c": {"old": None, "new": 3}},
        ),
    ],
)
def test_get_change_dict(exist, new, fields, expected):
    assert NATSApp._get_change_dict(exist, new, fields) == expected


@pytest.mark.asyncio
async def test_streams_create_or_update_add_stream(mocker):
    nc = NATSApp(url=["nats://localhost:4222"])
    config = StreamConfig(name="test_stream")
    nc._jetstream_configs.append(config)

    mock_js = mocker.Mock(spec=JetStreamContext)
    mock_js.stream_info.side_effect = NotFoundError
    mock_js.add_stream.return_value = mocker.Mock(as_dict=lambda: {"name": "test_stream"})
    nc._js = mock_js

    await nc._streams_create_or_update()

    mock_js.add_stream.assert_called_once_with(config)
    mock_js.stream_info.assert_called_once_with("test_stream")


@pytest.mark.asyncio
async def test_streams_create_or_update_no_change(mocker):
    nc = NATSApp(url=["nats://localhost:4222"])
    config = StreamConfig(name="test_stream", subjects=["foo"])
    nc._jetstream_configs.append(config)

    mock_js = mocker.Mock(spec=JetStreamContext)
    mock_js.stream_info.return_value = mocker.Mock(config=config)
    nc._js = mock_js

    mocker.patch.object(NATSApp, "_is_equal", return_value=True)

    await nc._streams_create_or_update()

    mock_js.update_stream.assert_not_called()
    mock_js.add_stream.assert_not_called()
    mock_js.stream_info.assert_called_once_with("test_stream")


@pytest.mark.asyncio
async def test_streams_create_or_update_deny_change(mocker):
    nc = NATSApp(url=["nats://localhost:4222"])
    config = StreamConfig(name="test_stream", subjects=["foo"])
    nc._jetstream_configs.append(config)

    mock_js = mocker.Mock(spec=JetStreamContext)
    mock_js.stream_info.return_value = mocker.Mock(config=config)
    nc._js = mock_js

    mocker.patch.object(NATSApp, "_is_equal", side_effect=[False, True])
    mocker.patch.object(NATSApp, "_get_change_dict", return_value={"name": {"old": "test_stream", "new": "new_stream"}})

    with pytest.raises(ValueError, match="nats: stream config params .* deny change"):
        await nc._streams_create_or_update()

    mock_js.update_stream.assert_not_called()
    mock_js.add_stream.assert_not_called()
    mock_js.stream_info.assert_called_once_with("test_stream")



import asyncio
from collections import defaultdict
from typing import Any, Optional

from nats.aio.msg import Msg

from nats_app.middlewares.validation import validate_args


async def test_rpc_call(nc):
    @nc.push_subscribe("app.subject.echo", queue="worker")
    async def rpc_func_echo(m: Msg) -> str:
        return f"Response: {m.data}"

    await nc.connect()

    res = await nc.request("app.subject.echo", "test")
    assert isinstance(res, str)
    assert res == "Response: b'\"test\"'"


async def test_rpc_call_w_validate(nc):
    @nc.push_subscribe("app.subject.validate_echo", queue="worker")
    @validate_args(by_pass_msg="msg")
    async def rpc_func_echo(data: str, msg: Any) -> str:
        return f"Response: {data} Msg.data: {msg.data}"

    await nc.connect()

    res = await nc.request("app.subject.validate_echo", {"args": ["test"]})
    assert isinstance(res, str)
    assert res == 'Response: test Msg.data: b\'{"args":["test"]}\''


async def test_rpc_call_w_validate_wo_param(nc):
    @nc.push_subscribe("app.subject.validate_echo", queue="worker")
    @validate_args
    async def rpc_func_echo(data: str) -> str:
        return f"Response: {data}"

    await nc.connect()

    res = await nc.request("app.subject.validate_echo", {"args": ["test"]})
    assert isinstance(res, str)
    assert res == "Response: test"


async def test_rpc_call_invalid_args(nc):
    @nc.push_subscribe("app.subject.validate_echo", queue="worker")
    @validate_args(by_pass_msg="msg")
    async def rpc_func_echo(data: str, val: int, key: bool, var1: Optional[float], var2: float, msg: Any) -> str:
        return f"Response: {data} Msg.data: {msg.data}"

    await nc.connect()

    res = await nc.request(
        "app.subject.validate_echo",
        {"args": ["test", "wow"], "kwargs": {"var1": "oto", "var2": 3.14}},
    )
    assert isinstance(res, dict)
    assert res == {
        "errors": [
            {
                "loc": [1],
                "msg": "Input should be a valid integer, unable to parse string as an integer",
                "type": "int_parsing",
            },
            {
                "loc": ["key"],
                "msg": "Missing required argument",
                "type": "missing_argument",
            },
            {
                "loc": ["var1"],
                "msg": "Input should be a valid number, unable to parse string as a number",
                "type": "float_parsing",
            },
        ]
    }


async def test_publish_w_validate(nc):
    result = {}

    @nc.push_subscribe("app.subject.pub")
    @validate_args()
    async def rpc_func_echo(data: str, data1: int) -> None:
        result["data"] = data
        result["data1"] = data1

    await nc.connect()

    await nc.publish("app.subject.pub", {"args": ["test"], "kwargs": {"data1": 1}})
    assert result["data"] == "test"
    assert result["data1"] == 1


async def test_publish_wo_validate(nc):
    result = {}

    @nc.push_subscribe("app.subject.pub_raw")
    async def rpc_func_echo(m: Msg) -> None:
        result["data"] = m.data

    await nc.connect()

    await nc.publish("app.subject.pub_raw", {"args": ["test"], "kwargs": {"data1": 1}})
    assert result["data"] == b'{"args":["test"],"kwargs":{"data1":1}}'


async def test_publish_w_queue(nc):
    c = defaultdict(int)
    result = {}

    @nc.push_subscribe("app.subject.pub_raw", queue="worker")
    async def rpc_func_echo(m: Msg) -> None:
        c["call"] += 1
        result["data"] = m.data

    @nc.push_subscribe("app.subject.pub_raw", queue="worker")
    async def rpc_func_echo_v2(m: Msg) -> None:
        c["call"] += 1
        result["data"] = m.data

    await nc.connect()

    await nc.publish("app.subject.pub_raw", {"args": ["test"], "kwargs": {"data1": 1}})
    assert result["data"] == b'{"args":["test"],"kwargs":{"data1":1}}'
    assert c["call"] == 1


async def test_publish_wo_queue(nc):
    c = defaultdict(int)
    result = {}

    @nc.push_subscribe("app.subject.pub_raw")
    async def rpc_func_echo(m: Msg) -> None:
        c["call"] += 1
        result["data"] = m.data

    @nc.push_subscribe("app.subject.pub_raw")
    async def rpc_func_echo_v2(m: Msg) -> None:
        c["call"] += 1
        result["data"] = m.data

    await nc.connect()

    await nc.publish("app.subject.pub_raw", {"args": ["test"], "kwargs": {"data1": 1}})
    assert result["data"] == b'{"args":["test"],"kwargs":{"data1":1}}'
    assert c["call"] == 2


async def test_push_subscription(nc):
    result = None

    @nc.js_push_subscribe("app.subject.validate_echo", queue="worker")
    async def handler(msg: Msg):
        nonlocal result
        result = msg.data
        await msg.ack()

    await nc.connect()

    await nc.js.publish("app.subject.validate_echo", b"TEST123")
    assert isinstance(result, bytes)
    assert result == b"TEST123"


async def test_pull_subscription(nc):
    result = None

    @nc.js_pull_subscribe("app.subject.validate_echo", batch=1)
    async def handler(msgs: list[Msg]):
        assert isinstance(msgs, list)
        assert len(msgs) == 1
        nonlocal result
        result = msgs[0].data

    await nc.connect()
    await nc.js.publish("app.subject.validate_echo", b"TEST123")
    assert isinstance(result, bytes)
    assert result == b"TEST123"



import logging

from fastapi.encoders import jsonable_encoder
from nats.aio.msg import Msg
from pydantic import ValidationError

try:
    from sentry_sdk import capture_exception
except ImportError:
    capture_exception = None


logger = logging.getLogger(__name__)


def default_error_handler(msg: Msg, e: Exception):
    logger.exception(f"got exception during handle request on subject: '{msg.subject}'")
    if isinstance(e, ValidationError):
        return {"errors": jsonable_encoder(e.errors(), exclude={"input", "url"})}

    if capture_exception:
        capture_exception(e)

    tb = e.__traceback__
    while tb.tb_next:
        tb = tb.tb_next
    kwargs = {
        "detail": str(e),
        "func": tb.tb_frame.f_code.co_name,
        "file": tb.tb_frame.f_code.co_filename.split("/")[-1],
        # "line": str(tb.tb_lineno),
    }

    return kwargs
